from typing import Union

import requests

GITLAB_URL = "https://code.ihep.ac.cn"
PROJECT_ID = "5548"

# API URL to get all tags of a project
TAGS_API_URL = f"{GITLAB_URL}/api/v4/projects/{PROJECT_ID}/repository/tags"


def get_all_tags():
    """
    Get all tags from git repository
    """
    tags = []
    page = 1
    while True:
        response = requests.get(TAGS_API_URL, params={"page": page, "per_page": 100})
        if response.status_code != 200:
            print(f"Error fetching tags: {response.status_code} - {response.text}")
            break

        data = response.json()
        if not data:
            break  # No more tags to fetch

        tags.extend(data)
        page += 1

    return [t["name"] for t in tags]


def get_latest_version(tags: list[str]) -> Union[tuple[int, int, int, int], None]:
    """
    Get the latest version from given tags. Only parse tags "vx.x.x" or "vx.x.x.x"

    Parameters:
        tags (list[str]): Tags lists

    Returns:
        Tuple of version number, or None if tags are invalid to be parsed.
    """
    latest_version = None

    for tag in tags:
        if not tag.startswith("v"):
            continue

        try:
            version = [int(i) for i in tag[1:].split(".")]
            if len(version) == 3:
                version.append(0)
            elif len(version) != 4:
                continue

            if latest_version is None or version > latest_version:
                latest_version = version
        except:
            continue

    return tuple(latest_version) if latest_version is not None else None


def check_new_version(test_version_tuple: Union[tuple[int, int, int, int], None] = None):
    """
    Check whether a new version is released. Print a warning if a newer version exists.

    Parameters:
        test_version_tuple: Tuple of version number, or None if no valid version is found.
    """
    tags = get_all_tags()
    latest_version = get_latest_version(tags)

    if latest_version is None:
        if test_version_tuple is None:  # Not being tested, keep silent
            return
        else:
            raise ValueError("Failed to get latest version from GitLab")

    if test_version_tuple is None:
        # get current version
        try:
            from ._version import version_tuple as current_version_tuple
        except ImportError:
            return
    else:
        current_version_tuple = test_version_tuple

    if len(current_version_tuple) > 4:  # dev version
        return

    if len(current_version_tuple) == 3:
        current_version = current_version_tuple + (0,)
    elif len(current_version_tuple) == 4:
        current_version = current_version_tuple
    else:
        raise ValueError("Invalid version tuple")

    if current_version < latest_version:
        new_version = (
            ".".join(map(str, latest_version))
            if latest_version[3] != 0
            else ".".join(map(str, latest_version[:3]))
        )
        print(
            f"Warning: A newer version is available: v{new_version}. "
            "It is recommended to upgrade to latest version, since `pybes3` is still in beta version "
            "and there might be bug fixes and new features in the latest version."
        )
        print(
            f"You can run `pip install pybes3 --upgrade [--user]` to upgrade to the latest version."
        )



from . import besio, detectors
from ._check_latest_version import check_new_version
from ._version import __version__, version
from .besio import concatenate, concatenate_raw, open, open_raw, wrap_uproot
from .detectors import (
    get_cgem_id,
    get_emc_id,
    get_mdc_id,
    get_muc_id,
    get_tof_mrpc_id,
    get_tof_scint_id,
    parse_cgem_id,
    parse_emc_id,
    parse_mdc_id,
    parse_muc_id,
    parse_tof_id,
)

check_new_version()



from typing import Literal, Union

import awkward as ak
import numba as nb
import numpy as np

DIGI_MDC_FLAG = np.uint32(0x10)
DIGI_TOF_FLAG = np.uint32(0x20)
DIGI_EMC_FLAG = np.uint32(0x30)
DIGI_MUC_FLAG = np.uint32(0x40)
DIGI_HLT_FLAG = np.uint32(0x50)
DIGI_CGEM_FLAG = np.uint32(0x60)
DIGI_MRPC_FLAG = np.uint32(0x70)
DIGI_FLAG_OFFSET = np.uint32(24)
DIGI_FLAG_MASK = np.uint32(0xFF000000)

# MDC
DIGI_MDC_WIRETYPE_OFFSET = np.uint32(15)
DIGI_MDC_WIRETYPE_MASK = np.uint32(0x00008000)
DIGI_MDC_LAYER_OFFSET = np.uint32(9)
DIGI_MDC_LAYER_MASK = np.uint32(0x00007E00)
DIGI_MDC_WIRE_OFFSET = np.uint32(0)
DIGI_MDC_WIRE_MASK = np.uint32(0x000001FF)
DIGI_MDC_STEREO_WIRE = np.uint32(1)

# TOF
DIGI_TOF_PART_OFFSET = np.uint32(14)
DIGI_TOF_PART_MASK = np.uint32(0x0000C000)
DIGI_TOF_END_OFFSET = np.uint32(0)
DIGI_TOF_END_MASK = np.uint32(0x00000001)

DIGI_TOF_SCINT_LAYER_OFFSET = np.uint32(8)
DIGI_TOF_SCINT_LAYER_MASK = np.uint32(0x00000100)
DIGI_TOF_SCINT_PHI_OFFSET = np.uint32(1)
DIGI_TOF_SCINT_PHI_MASK = np.uint32(0x000000FE)

DIGI_TOF_MRPC_ENDCAP_OFFSET = np.uint32(11)
DIGI_TOF_MRPC_ENDCAP_MASK = np.uint32(0x00000800)
DIGI_TOF_MRPC_MODULE_OFFSET = np.uint32(5)
DIGI_TOF_MRPC_MODULE_MASK = np.uint32(0x000007E0)
DIGI_TOF_MRPC_STRIP_OFFSET = np.uint32(1)
DIGI_TOF_MRPC_STRIP_MASK = np.uint32(0x0000001E)

# EMC
DIGI_EMC_MODULE_OFFSET = np.uint32(16)
DIGI_EMC_MODULE_MASK = np.uint32(0x000F0000)
DIGI_EMC_THETA_OFFSET = np.uint32(8)
DIGI_EMC_THETA_MASK = np.uint32(0x00003F00)
DIGI_EMC_PHI_OFFSET = np.uint32(0)
DIGI_EMC_PHI_MASK = np.uint32(0x000000FF)

# MUC
DIGI_MUC_PART_OFFSET = np.uint32(16)
DIGI_MUC_PART_MASK = np.uint32(0x000F0000)
DIGI_MUC_SEGMENT_OFFSET = np.uint32(12)
DIGI_MUC_SEGMENT_MASK = np.uint32(0x0000F000)
DIGI_MUC_LAYER_OFFSET = np.uint32(8)
DIGI_MUC_LAYER_MASK = np.uint32(0x00000F00)
DIGI_MUC_CHANNEL_OFFSET = np.uint32(0)
DIGI_MUC_CHANNEL_MASK = np.uint32(0x000000FF)

# CGEM
DIGI_CGEM_STRIP_OFFSET = np.uint32(7)
DIGI_CGEM_STRIP_MASK = np.uint32(0x0007FF80)
DIGI_CGEM_STRIPTYPE_OFFSET = np.uint32(6)
DIGI_CGEM_STRIPTYPE_MASK = np.uint32(0x00000040)
DIGI_CGEM_SHEET_OFFSET = np.uint32(3)
DIGI_CGEM_SHEET_MASK = np.uint32(0x00000038)
DIGI_CGEM_LAYER_OFFSET = np.uint32(0)
DIGI_CGEM_LAYER_MASK = np.uint32(0x00000007)
DIGI_CGEM_XSTRIP = np.uint32(0)


###############################################################################
#                                     MDC                                     #
###############################################################################
@nb.vectorize([nb.bool(nb.int_)])
def check_mdc_id(
    mdc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.bool]:
    """
    Check if the MDC ID is valid.

    Parameters:
        mdc_id: The MDC ID array or value.

    Returns:
        Whether the ID is valid.
    """
    return (mdc_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_MDC_FLAG


@nb.vectorize([nb.uint16(nb.int_)])
def mdc_id_to_wire(
    mdc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint16]:
    """
    Convert MDC ID to wire ID

    Parameters:
        mdc_id: MDC ID array or value.

    Returns:
        The wire ID.
    """
    return (mdc_id & DIGI_MDC_WIRE_MASK) >> DIGI_MDC_WIRE_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def mdc_id_to_layer(
    mdc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Convert the MDC ID to the layer ID.

    Parameters:
        mdc_id: The MDC ID array or value.

    Returns:
        The layer ID.
    """
    return (mdc_id & DIGI_MDC_LAYER_MASK) >> DIGI_MDC_LAYER_OFFSET


@nb.vectorize([nb.bool(nb.int_)])
def mdc_id_to_is_stereo(
    mdc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.bool]:
    """
    Convert the MDC ID to whether it is a stereo wire.

    Parameters:
        mdc_id: The MDC ID array or value.

    Returns:
        Whether the wire is a stereo wire.
    """
    return (
        mdc_id & DIGI_MDC_WIRETYPE_MASK
    ) >> DIGI_MDC_WIRETYPE_OFFSET == DIGI_MDC_STEREO_WIRE


@nb.vectorize([nb.uint32(nb.int_, nb.int_, nb.int_), nb.uint32(nb.int_, nb.int_, nb.bool)])
def get_mdc_id(
    wire: Union[ak.Array, np.ndarray, int],
    layer: Union[ak.Array, np.ndarray, int],
    wire_type: Union[ak.Array, np.ndarray, int],
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Generate MDC ID based on the wire ID, layer ID, and wire type.

    Parameters:
        wire: The wire ID.
        layer: The layer ID.
        wire_type: The wire type.

    Returns:
        The MDC ID.
    """
    return (
        ((wire << DIGI_MDC_WIRE_OFFSET) & DIGI_MDC_WIRE_MASK)
        | ((layer << DIGI_MDC_LAYER_OFFSET) & DIGI_MDC_LAYER_MASK)
        | ((wire_type << DIGI_MDC_WIRETYPE_OFFSET) & DIGI_MDC_WIRETYPE_MASK)
        | (DIGI_MDC_FLAG << DIGI_FLAG_OFFSET)
    )


###############################################################################
#                                     TOF                                     #
###############################################################################
@nb.vectorize([nb.bool(nb.int_)])
def check_tof_id(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.bool]:
    """
    Check if the TOF ID is valid.

    Parameters:
        tof_id: The TOF ID array or value.

    Returns:
        Whether the ID is valid.
    """
    return (tof_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_TOF_FLAG


@nb.vectorize([nb.uint8(nb.int_)])
def tof_id_to_part(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert TOF ID to part ID

    Parameters:
        tof_id: TOF ID array or value.

    Returns:
        The part ID.
    """
    return (tof_id & DIGI_TOF_PART_MASK) >> DIGI_TOF_PART_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def tof_id_to_end(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the TOF ID to the readout end ID.

    Parameters:
        tof_id: The TOF ID array or value.

    Returns:
        The readout end ID.
    """
    return (tof_id & DIGI_TOF_END_MASK) >> DIGI_TOF_END_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def tof_id_to_scint_layer(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the TOF ID to the scintillator layer ID.

    Parameters:
        tof_id: The TOF ID array or value.

    Returns:
        The scintillator layer ID.
    """
    return (tof_id & DIGI_TOF_SCINT_LAYER_MASK) >> DIGI_TOF_SCINT_LAYER_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def tof_id_to_scint_phi(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the TOF ID to the scintillator phi ID.

    Parameters:
        tof_id: The TOF ID array or value.

    Returns:
        The scintillator phi ID.
    """
    return (tof_id & DIGI_TOF_SCINT_PHI_MASK) >> DIGI_TOF_SCINT_PHI_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def tof_id_to_mrpc_endcap(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the TOF ID to the MRPC endcap ID.

    Parameters:
        tof_id: The TOF ID array or value.

    Returns:
        The MRPC endcap ID.
    """
    return (tof_id & DIGI_TOF_MRPC_ENDCAP_MASK) >> DIGI_TOF_MRPC_ENDCAP_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def tof_id_to_mrpc_module(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the TOF ID to the MRPC module ID.

    Parameters:
        tof_id: The TOF ID array or value.

    Returns:
        The MRPC module ID.
    """
    return (tof_id & DIGI_TOF_MRPC_MODULE_MASK) >> DIGI_TOF_MRPC_MODULE_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def tof_id_to_mrpc_strip(
    tof_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the TOF ID to the MRPC strip ID.

    Parameters:
        tof_id: The TOF ID array or value.

    Returns:
        The MRPC strip ID.
    """
    return (tof_id & DIGI_TOF_MRPC_STRIP_MASK) >> DIGI_TOF_MRPC_STRIP_OFFSET


@nb.vectorize([nb.uint32(nb.int_, nb.int_, nb.int_, nb.int_)])
def get_tof_scint_id(
    part: Union[ak.Array, np.ndarray, int],
    layer: Union[ak.Array, np.ndarray, int],
    phi: Union[ak.Array, np.ndarray, int],
    end: Union[ak.Array, np.ndarray, int],
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Generate TOF scintillator ID based on the part ID, layer ID, phi ID, and readout end ID.

    Parameters:
        part: The part ID.
        layer: The layer ID.
        phi: The phi ID.
        end: The readout end ID.

    Returns:
        The TOF scintillator ID.
    """
    return (
        ((part << DIGI_TOF_PART_OFFSET) & DIGI_TOF_PART_MASK)
        | ((layer << DIGI_TOF_SCINT_LAYER_OFFSET) & DIGI_TOF_SCINT_LAYER_MASK)
        | ((phi << DIGI_TOF_SCINT_PHI_OFFSET) & DIGI_TOF_SCINT_PHI_MASK)
        | ((end << DIGI_TOF_END_OFFSET) & DIGI_TOF_END_MASK)
        | (DIGI_TOF_FLAG << DIGI_FLAG_OFFSET)
    )


@nb.vectorize([nb.uint32(nb.int_, nb.int_, nb.int_, nb.int_)])
def get_tof_mrpc_id(
    endcap: Union[ak.Array, np.ndarray, int],
    module: Union[ak.Array, np.ndarray, int],
    strip: Union[ak.Array, np.ndarray, int],
    end: Union[ak.Array, np.ndarray, int],
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Generate TOF MRPC ID based on the endcap ID, module ID, strip ID, and readout end ID.

    Parameters:
        endcap: The endcap ID.
        module: The module ID.
        strip: The strip ID.
        end: The readout end ID.

    Returns:
        The TOF MRPC ID.
    """
    return (
        ((np.uint32(3) << DIGI_TOF_PART_OFFSET) & DIGI_TOF_PART_MASK)
        | ((endcap << DIGI_TOF_MRPC_ENDCAP_OFFSET) & DIGI_TOF_MRPC_ENDCAP_MASK)
        | ((module << DIGI_TOF_MRPC_MODULE_OFFSET) & DIGI_TOF_MRPC_MODULE_MASK)
        | ((strip << DIGI_TOF_MRPC_STRIP_OFFSET) & DIGI_TOF_MRPC_STRIP_MASK)
        | ((end << DIGI_TOF_END_OFFSET) & DIGI_TOF_END_MASK)
        | (DIGI_TOF_FLAG << DIGI_FLAG_OFFSET)
    )


###############################################################################
#                                     EMC                                     #
###############################################################################
@nb.vectorize([nb.bool(nb.int_)])
def check_emc_id(
    emc_id: Union[ak.Array, np.ndarray, np.uint32]
) -> Union[ak.Array, np.ndarray, np.bool]:
    """
    Check if the EMC ID is valid.

    Parameters:
        emc_id: The EMC ID array or value.

    Returns:
        Whether the ID is valid.
    """
    return (emc_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_EMC_FLAG


@nb.vectorize([nb.uint8(nb.int_)])
def emc_id_to_module(
    emc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert EMC ID to module ID

    Parameters:
        emc_id: EMC ID array or value.

    Returns:
        The module ID.
    """
    return (emc_id & DIGI_EMC_MODULE_MASK) >> DIGI_EMC_MODULE_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def emc_id_to_theta(
    emc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the EMC ID to the theta ID.

    Parameters:
        emc_id: The EMC ID array or value.

    Returns:
        The theta ID.
    """
    return (emc_id & DIGI_EMC_THETA_MASK) >> DIGI_EMC_THETA_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def emc_id_to_phi(
    emc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the EMC ID to the phi ID.

    Parameters:
        emc_id: The EMC ID array or value.

    Returns:
        The phi ID.
    """
    return (emc_id & DIGI_EMC_PHI_MASK) >> DIGI_EMC_PHI_OFFSET


@nb.vectorize([nb.uint32(nb.int_, nb.int_, nb.int_)])
def get_emc_id(
    module: Union[ak.Array, np.ndarray, int],
    theta: Union[ak.Array, np.ndarray, int],
    phi: Union[ak.Array, np.ndarray, int],
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Generate EMC ID based on the module ID, theta ID, and phi ID.

    Parameters:
        module: The module ID.
        theta: The theta ID.
        phi: The phi ID.

    Returns:
        The EMC ID.
    """
    return (
        ((module << DIGI_EMC_MODULE_OFFSET) & DIGI_EMC_MODULE_MASK)
        | ((theta << DIGI_EMC_THETA_OFFSET) & DIGI_EMC_THETA_MASK)
        | ((phi << DIGI_EMC_PHI_OFFSET) & DIGI_EMC_PHI_MASK)
        | (DIGI_EMC_FLAG << DIGI_FLAG_OFFSET)
    )


###############################################################################
#                                     MUC                                     #
###############################################################################
@nb.vectorize([nb.bool(nb.int_)])
def check_muc_id(
    muc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.bool]:
    """
    Check if the MUC ID is valid.

    Parameters:
        muc_id: The MUC ID array or value.

    Returns:
        Whether the ID is valid.
    """
    return (muc_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_MUC_FLAG


@nb.vectorize([nb.uint8(nb.int_)])
def muc_id_to_part(
    muc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert MUC ID to part ID

    Parameters:
        muc_id: MUC ID array or value.

    Returns:
        The part ID.
    """
    return (muc_id & DIGI_MUC_PART_MASK) >> DIGI_MUC_PART_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def muc_id_to_segment(
    muc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the MUC ID to the segment ID.

    Parameters:
        muc_id: The MUC ID array or value.

    Returns:
        The segment ID.
    """
    return (muc_id & DIGI_MUC_SEGMENT_MASK) >> DIGI_MUC_SEGMENT_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def muc_id_to_layer(
    muc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the MUC ID to the layer ID.

    Parameters:
        muc_id: The MUC ID array or value.

    Returns:
        The layer ID.
    """
    return (muc_id & DIGI_MUC_LAYER_MASK) >> DIGI_MUC_LAYER_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def muc_id_to_channel(
    muc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the MUC ID to the channel ID.

    Parameters:
        muc_id: The MUC ID array or value.

    Returns:
        The channel ID.
    """
    return (muc_id & DIGI_MUC_CHANNEL_MASK) >> DIGI_MUC_CHANNEL_OFFSET


@nb.vectorize([nb.uint32(nb.int_, nb.int_, nb.int_, nb.int_)])
def get_muc_id(
    part: Union[ak.Array, np.ndarray, int],
    segment: Union[ak.Array, np.ndarray, int],
    layer: Union[ak.Array, np.ndarray, int],
    channel: Union[ak.Array, np.ndarray, int],
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Generate MUC ID based on the part ID, segment ID, layer ID, and channel ID.

    Parameters:
        part: The part ID.
        segment: The segment ID.
        layer: The layer ID.
        channel: The channel ID.

    Returns:
        The MUC ID.
    """
    return (
        ((part << DIGI_MUC_PART_OFFSET) & DIGI_MUC_PART_MASK)
        | ((segment << DIGI_MUC_SEGMENT_OFFSET) & DIGI_MUC_SEGMENT_MASK)
        | ((layer << DIGI_MUC_LAYER_OFFSET) & DIGI_MUC_LAYER_MASK)
        | ((channel << DIGI_MUC_CHANNEL_OFFSET) & DIGI_MUC_CHANNEL_MASK)
        | (DIGI_MUC_FLAG << DIGI_FLAG_OFFSET)
    )


def muc_id_to_gap(
    muc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the MUC ID to the gap ID, which is equivalent to layer ID.

    Parameters:
        muc_id: The MUC ID array or value.

    Returns:
        The gap ID.
    """
    return muc_id_to_layer(muc_id)


def muc_id_to_strip(
    muc_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the MUC ID to the strip ID, which is equivalent to channel ID.

    Parameters:
        muc_id: The MUC ID array or value.

    Returns:
        The strip ID.
    """
    return muc_id_to_channel(muc_id)


###############################################################################
#                                    CGEM                                     #
###############################################################################
@nb.vectorize([nb.bool(nb.int_)])
def check_cgem_id(
    cgem_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.bool]:
    """
    Check if the CGEM ID is valid.

    Parameters:
        cgem_id: The CGEM ID array or value.

    Returns:
        Whether the ID is valid.
    """
    return (cgem_id & DIGI_FLAG_MASK) >> DIGI_FLAG_OFFSET == DIGI_CGEM_FLAG


@nb.vectorize([nb.uint8(nb.int_)])
def cgem_id_to_layer(
    cgem_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the CGEM ID to the layer ID.

    Parameters:
        cgem_id: The CGEM ID array or value.

    Returns:
        The layer ID.
    """
    return (cgem_id & DIGI_CGEM_LAYER_MASK) >> DIGI_CGEM_LAYER_OFFSET


@nb.vectorize([nb.uint8(nb.int_)])
def cgem_id_to_sheet(
    cgem_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint8]:
    """
    Convert the CGEM ID to the sheet ID.

    Parameters:
        cgem_id: The CGEM ID array or value.

    Returns:
        The sheet ID.
    """
    return (cgem_id & DIGI_CGEM_SHEET_MASK) >> DIGI_CGEM_SHEET_OFFSET


@nb.vectorize([nb.uint16(nb.int_)])
def cgem_id_to_strip(
    cgem_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Convert CGEM ID to strip ID

    Parameters:
        cgem_id: CGEM ID array or value.

    Returns:
        The strip ID.
    """
    return (cgem_id & DIGI_CGEM_STRIP_MASK) >> DIGI_CGEM_STRIP_OFFSET


@nb.vectorize([nb.bool(nb.int_)])
def cgem_id_to_is_x_strip(
    cgem_id: Union[ak.Array, np.ndarray, int]
) -> Union[ak.Array, np.ndarray, np.bool]:
    """
    Convert the CGEM ID to whether it is an X-strip.

    Parameters:
        cgem_id: The CGEM ID array or value.

    Returns:
        Whether the strip is an X-strip
    """
    return (
        (cgem_id & DIGI_CGEM_STRIPTYPE_MASK) >> DIGI_CGEM_STRIPTYPE_OFFSET
    ) == DIGI_CGEM_XSTRIP


@nb.vectorize([nb.uint32(nb.int_, nb.int_, nb.int_, nb.bool)])
def get_cgem_id(
    layer: Union[ak.Array, np.ndarray, int],
    sheet: Union[ak.Array, np.ndarray, int],
    strip: Union[ak.Array, np.ndarray, int],
    is_x_strip: Union[ak.Array, np.ndarray, bool],
) -> Union[ak.Array, np.ndarray, np.uint32]:
    """
    Generate CGEM ID based on the strip ID, strip type, sheet ID, and layer ID.

    Parameters:
        layer: The layer ID.
        sheet: The sheet ID.
        strip: The strip ID.
        is_x_strip: Whether the strip is an X-strip.

    Returns:
        The CGEM ID.
    """
    return (
        ((strip << DIGI_CGEM_STRIP_OFFSET) & DIGI_CGEM_STRIP_MASK)
        | ((~is_x_strip << DIGI_CGEM_STRIPTYPE_OFFSET) & DIGI_CGEM_STRIPTYPE_MASK)
        | ((sheet << DIGI_CGEM_SHEET_OFFSET) & DIGI_CGEM_SHEET_MASK)
        | ((layer << DIGI_CGEM_LAYER_OFFSET) & DIGI_CGEM_LAYER_MASK)
        | (DIGI_CGEM_FLAG << DIGI_FLAG_OFFSET)
    )


###############################################################################
#                                Make awkwards                                #
###############################################################################
def parse_mdc_id(
    mdc_id: Union[ak.Array, np.ndarray, int],
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse MDC ID.

    If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:
        - wire: The wire ID.
        - layer: The layer ID.
        - is_stereo: Whether the wire is a stereo wire.

    Parameters:
        mdc_id: The MDC ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed MDC ID. If `library` is `ak`, return `ak.Record`.
    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(mdc_id, ak.Array):
        mdc_id = ak.flatten(mdc_id)

    res = {
        "wire": mdc_id_to_wire(mdc_id),
        "layer": mdc_id_to_layer(mdc_id),
        "is_stereo": mdc_id_to_is_stereo(mdc_id),
    }

    if library == "ak":
        return ak.zip(res)
    else:
        return res


def parse_tof_id(
    tof_id: Union[ak.Array, np.ndarray, int],
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse TOF ID.

    If `library` is `ak`, return `ak.Record`. If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:
        - part: The part ID.
        - end: The readout end ID.
        - scint_layer: The scintillator layer ID.
        - scint_phi: The scintillator phi ID.
        - mrpc_endcap: The MRPC endcap ID.
        - mrpc_module: The MRPC module ID.
        - mrpc_strip: The MRPC strip ID.
        - is_mrpc: Whether the TOF ID is MRPC ID.

    Note that no matter an ID is MRPC or not, it will always have both scintillator and MRPC keys.

    Parameters:
        tof_id: The TOF ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed TOF ID.

    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(tof_id, ak.Array):
        tof_id = ak.flatten(tof_id)

    res = {
        "part": tof_id_to_part(tof_id),
        "end": tof_id_to_end(tof_id),
        "scint_layer": tof_id_to_scint_layer(tof_id),
        "scint_phi": tof_id_to_scint_phi(tof_id),
        "mrpc_endcap": tof_id_to_mrpc_endcap(tof_id),
        "mrpc_module": tof_id_to_mrpc_module(tof_id),
        "mrpc_strip": tof_id_to_mrpc_strip(tof_id),
    }

    res["is_mrpc"] = res["part"] == 3

    if library == "ak":
        return ak.zip(res)
    else:
        return res


def parse_emc_id(
    emc_id: Union[ak.Array, np.ndarray, int],
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse EMC ID.

    If `library` is `ak`, return `ak.Record`. If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:
        - module: The module ID.
        - theta: The theta ID.
        - phi: The phi ID.

    Parameters:
        emc_id: The EMC ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed EMC ID.

    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(emc_id, ak.Array):
        emc_id = ak.flatten(emc_id)

    res = {
        "module": emc_id_to_module(emc_id),
        "theta": emc_id_to_theta(emc_id),
        "phi": emc_id_to_phi(emc_id),
    }

    if library == "ak":
        return ak.zip(res)
    else:
        return res


def parse_muc_id(
    muc_id: Union[ak.Array, np.ndarray, int],
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse MUC ID.

    If `library` is `ak`, return `ak.Record`. If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:
        - part: The part ID.
        - segment: The segment ID.
        - layer: The layer ID.
        - channel: The channel ID.
        - gap: The gap ID, which is equivalent to layer ID.
        - strip: The strip ID, which is equivalent to channel ID.

    Parameters:
        muc_id: The MUC ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed MUC ID.
    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(muc_id, ak.Array):
        muc_id = ak.flatten(muc_id)

    part = muc_id_to_part(muc_id)
    segment = muc_id_to_segment(muc_id)
    layer = muc_id_to_layer(muc_id)
    channel = muc_id_to_channel(muc_id)

    res = {
        "part": part,
        "segment": segment,
        "layer": layer,
        "channel": channel,
        "gap": layer,
        "strip": channel,
    }

    if library == "ak":
        return ak.zip(res)
    else:
        return res


def parse_cgem_id(
    cgem_id: Union[ak.Array, np.ndarray, int],
    flat: bool = False,
    library: Literal["ak", "np"] = "ak",
) -> Union[ak.Record, dict[str, np.ndarray], dict[str, np.int_]]:
    """
    Parse CGEM ID.

    If `library` is `ak`, return `ak.Record`. If `library` is `np`, return `dict[str, np.ndarray]`.

    Available keys of the output:
        - layer: The layer ID.
        - sheet: The sheet ID.
        - strip: The strip ID.
        - is_x_strip: Whether the strip is an X-strip.

    Parameters:
        cgem_id: The CGEM ID.
        flat: Whether to flatten the output.
        library: The library to use as output.

    Returns:
        The parsed CGEM ID.
    """
    if library not in ["ak", "np"]:
        raise ValueError(f"Unsupported library: {library}")

    if flat and isinstance(cgem_id, ak.Array):
        cgem_id = ak.flatten(cgem_id)

    res = {
        "layer": cgem_id_to_layer(cgem_id),
        "sheet": cgem_id_to_sheet(cgem_id),
        "strip": cgem_id_to_strip(cgem_id),
        "is_x_strip": cgem_id_to_is_x_strip(cgem_id),
    }

    if library == "ak":
        return ak.zip(res)
    else:
        return res



from pybes3._check_latest_version import check_new_version, get_all_tags, get_latest_version


def test_get_all_tags():
    tags = get_all_tags()
    assert len(tags) > 0


def test_get_latest_version():
    tags1 = ["v0.1.0", "v0.1.1"]
    latest_version1 = get_latest_version(tags1)
    assert latest_version1 == (0, 1, 1, 0)

    tags2 = ["v0.1.0", "v0.1.0.1"]
    latest_version2 = get_latest_version(tags2)
    assert latest_version2 == (0, 1, 0, 1)

    tags3 = ["v0.1.0", "v0.1.0.1", "v0.1.0.2"]
    latest_version3 = get_latest_version(tags3)
    assert latest_version3 == (0, 1, 0, 2)

    tags4 = ["v0.1.0", "v0.1.0.1", "v0.1.0.2", "v0.2.0"]
    latest_version4 = get_latest_version(tags4)
    assert latest_version4 == (0, 2, 0, 0)

    # v0.3 should be ignored
    tags5 = ["v0.1.0", "v0.1.0.1", "v0.1.0.2", "v0.2.0", "v0.3"]
    latest_version5 = get_latest_version(tags5)
    assert latest_version5 == (0, 2, 0, 0)


def test_check_latest_version(capsys):
    test_current_version1 = (99, 99, 99, 99)
    check_new_version(test_current_version1)
    captured = capsys.readouterr()
    assert captured.out == ""

    test_current_version2 = (0, 1, 0, 1)
    check_new_version(test_current_version2)
    captured = capsys.readouterr()
    assert len(captured.out) > 0



from pathlib import Path

import awkward as ak
import numpy as np

import pybes3 as p3
from pybes3.detectors import identify as det_id

data_dir = Path(__file__).parent.parent / "data"


def test_mdc_id():
    mdc_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_mdcDigiCol"
    ].array()["m_intId"]

    # Test awkward
    ak_wire, ak_layer, ak_is_stereo = (
        det_id.mdc_id_to_wire(mdc_id_ak),
        det_id.mdc_id_to_layer(mdc_id_ak),
        det_id.mdc_id_to_is_stereo(mdc_id_ak),
    )
    assert ak.all(p3.get_mdc_id(ak_wire, ak_layer, ak_is_stereo) == mdc_id_ak)
    assert ak.all(det_id.check_mdc_id(mdc_id_ak))

    # Test numpy
    mdc_id_np = ak.flatten(mdc_id_ak).to_numpy()
    np_wire, np_layer, np_is_stereo = (
        det_id.mdc_id_to_wire(mdc_id_np),
        det_id.mdc_id_to_layer(mdc_id_np),
        det_id.mdc_id_to_is_stereo(mdc_id_np),
    )
    assert np.all(p3.get_mdc_id(np_wire, np_layer, np_is_stereo) == mdc_id_np)
    assert np.all(det_id.check_mdc_id(mdc_id_np))

    # Test uint32
    tmp_id = mdc_id_np[0]
    tmp_wire, tmp_layer, tmp_is_stereo = (
        det_id.mdc_id_to_wire(tmp_id),
        det_id.mdc_id_to_layer(tmp_id),
        det_id.mdc_id_to_is_stereo(tmp_id),
    )
    assert tmp_wire == np_wire[0]
    assert tmp_layer == np_layer[0]
    assert tmp_is_stereo == np_is_stereo[0]
    assert p3.get_mdc_id(tmp_wire, tmp_layer, tmp_is_stereo) == tmp_id
    assert det_id.check_mdc_id(tmp_id)

    # Test python int
    tmp_id = int(mdc_id_np[0])
    tmp_wire, tmp_layer, tmp_is_stereo = (
        det_id.mdc_id_to_wire(tmp_id),
        det_id.mdc_id_to_layer(tmp_id),
        det_id.mdc_id_to_is_stereo(tmp_id),
    )
    assert tmp_wire == np_wire[0]
    assert tmp_layer == np_layer[0]
    assert tmp_is_stereo == np_is_stereo[0]
    assert p3.get_mdc_id(tmp_wire, tmp_layer, tmp_is_stereo) == tmp_id
    assert det_id.check_mdc_id(tmp_id)


def test_emc_id():
    emc_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_emcDigiCol"
    ].array()["m_intId"]

    # Test awkward
    ak_module, ak_theta, ak_phi = (
        det_id.emc_id_to_module(emc_id_ak),
        det_id.emc_id_to_theta(emc_id_ak),
        det_id.emc_id_to_phi(emc_id_ak),
    )
    assert ak.all(p3.get_emc_id(ak_module, ak_theta, ak_phi) == emc_id_ak)
    assert ak.all(det_id.check_emc_id(emc_id_ak))

    # Test numpy
    emc_id_np = ak.flatten(emc_id_ak).to_numpy()
    np_module, np_theta, np_phi = (
        det_id.emc_id_to_module(emc_id_np),
        det_id.emc_id_to_theta(emc_id_np),
        det_id.emc_id_to_phi(emc_id_np),
    )
    assert np.all(p3.get_emc_id(np_module, np_theta, np_phi) == emc_id_np)
    assert np.all(det_id.check_emc_id(emc_id_np))

    # Test uint32
    tmp_id = emc_id_np[0]
    tmp_module, tmp_theta, tmp_phi = (
        det_id.emc_id_to_module(tmp_id),
        det_id.emc_id_to_theta(tmp_id),
        det_id.emc_id_to_phi(tmp_id),
    )
    assert tmp_module == np_module[0]
    assert tmp_theta == np_theta[0]
    assert tmp_phi == np_phi[0]
    assert p3.get_emc_id(tmp_module, tmp_theta, tmp_phi) == tmp_id
    assert det_id.check_emc_id(tmp_id)

    # Test python int
    tmp_id = int(emc_id_np[0])
    tmp_module, tmp_theta, tmp_phi = (
        det_id.emc_id_to_module(tmp_id),
        det_id.emc_id_to_theta(tmp_id),
        det_id.emc_id_to_phi(tmp_id),
    )
    assert tmp_module == np_module[0]
    assert tmp_theta == np_theta[0]
    assert tmp_phi == np_phi[0]
    assert p3.get_emc_id(tmp_module, tmp_theta, tmp_phi) == tmp_id
    assert det_id.check_emc_id(tmp_id)


def test_tof_scint_id():
    tof_scint_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_tofDigiCol"
    ].array()["m_intId"]

    # Test awkward
    ak_part, ak_layer, ak_phi, ak_end = (
        det_id.tof_id_to_part(tof_scint_id_ak),
        det_id.tof_id_to_scint_layer(tof_scint_id_ak),
        det_id.tof_id_to_scint_phi(tof_scint_id_ak),
        det_id.tof_id_to_end(tof_scint_id_ak),
    )
    assert ak.all(p3.get_tof_scint_id(ak_part, ak_layer, ak_phi, ak_end) == tof_scint_id_ak)
    assert ak.all(det_id.check_tof_id(tof_scint_id_ak))

    # Test numpy
    tof_scint_id_np = ak.flatten(tof_scint_id_ak).to_numpy()
    np_part, np_layer, np_phi, np_end = (
        det_id.tof_id_to_part(tof_scint_id_np),
        det_id.tof_id_to_scint_layer(tof_scint_id_np),
        det_id.tof_id_to_scint_phi(tof_scint_id_np),
        det_id.tof_id_to_end(tof_scint_id_np),
    )
    assert np.all(p3.get_tof_scint_id(np_part, np_layer, np_phi, np_end) == tof_scint_id_np)
    assert np.all(det_id.check_tof_id(tof_scint_id_np))

    # Test uint32
    tmp_id = tof_scint_id_np[0]
    tmp_part, tmp_layer, tmp_phi, tmp_end = (
        det_id.tof_id_to_part(tmp_id),
        det_id.tof_id_to_scint_layer(tmp_id),
        det_id.tof_id_to_scint_phi(tmp_id),
        det_id.tof_id_to_end(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_layer == np_layer[0]
    assert tmp_phi == np_phi[0]
    assert tmp_end == np_end[0]
    assert p3.get_tof_scint_id(tmp_part, tmp_layer, tmp_phi, tmp_end) == tmp_id
    assert det_id.check_tof_id(tmp_id)

    # Test python int
    tmp_id = int(tof_scint_id_np[0])
    tmp_part, tmp_layer, tmp_phi, tmp_end = (
        det_id.tof_id_to_part(tmp_id),
        det_id.tof_id_to_scint_layer(tmp_id),
        det_id.tof_id_to_scint_phi(tmp_id),
        det_id.tof_id_to_end(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_layer == np_layer[0]
    assert tmp_phi == np_phi[0]
    assert tmp_end == np_end[0]
    assert p3.get_tof_scint_id(tmp_part, tmp_layer, tmp_phi, tmp_end) == tmp_id
    assert det_id.check_tof_id(tmp_id)


def test_tof_mrpc_id():
    tof_mrpc_id_ak: ak.Array = p3.open(data_dir / "test_mrpc.rtraw")[
        "Event/TDigiEvent/m_tofDigiCol"
    ].array()["m_intId"]

    # Test awkward
    ak_endcap, ak_module, ak_strip, ak_end = (
        det_id.tof_id_to_mrpc_endcap(tof_mrpc_id_ak),
        det_id.tof_id_to_mrpc_module(tof_mrpc_id_ak),
        det_id.tof_id_to_mrpc_strip(tof_mrpc_id_ak),
        det_id.tof_id_to_end(tof_mrpc_id_ak),
    )

    is_mrpc = det_id.tof_id_to_part(tof_mrpc_id_ak) == 3
    assert ak.all(
        p3.get_tof_mrpc_id(ak_endcap, ak_module, ak_strip, ak_end)[is_mrpc]
        == tof_mrpc_id_ak[is_mrpc]
    )
    assert ak.all(det_id.check_tof_id(tof_mrpc_id_ak))

    # Test numpy
    tof_mrpc_id_np = ak.flatten(tof_mrpc_id_ak).to_numpy()
    np_endcap, np_module, np_strip, np_end = (
        det_id.tof_id_to_mrpc_endcap(tof_mrpc_id_np),
        det_id.tof_id_to_mrpc_module(tof_mrpc_id_np),
        det_id.tof_id_to_mrpc_strip(tof_mrpc_id_np),
        det_id.tof_id_to_end(tof_mrpc_id_np),
    )

    is_mrpc = det_id.tof_id_to_part(tof_mrpc_id_np) == 3
    assert np.all(
        p3.get_tof_mrpc_id(np_endcap, np_module, np_strip, np_end)[is_mrpc]
        == tof_mrpc_id_np[is_mrpc]
    )
    assert np.all(det_id.check_tof_id(tof_mrpc_id_np))

    # Test uint32
    is_mrpc = det_id.tof_id_to_part(tof_mrpc_id_np) == 3
    tmp_id = tof_mrpc_id_np[is_mrpc][0]
    tmp_endcap, tmp_module, tmp_strip, tmp_end = (
        det_id.tof_id_to_mrpc_endcap(tmp_id),
        det_id.tof_id_to_mrpc_module(tmp_id),
        det_id.tof_id_to_mrpc_strip(tmp_id),
        det_id.tof_id_to_end(tmp_id),
    )
    assert tmp_endcap == np_endcap[is_mrpc][0]
    assert tmp_module == np_module[is_mrpc][0]
    assert tmp_strip == np_strip[is_mrpc][0]
    assert tmp_end == np_end[is_mrpc][0]
    assert p3.get_tof_mrpc_id(tmp_endcap, tmp_module, tmp_strip, tmp_end) == tmp_id
    assert det_id.check_tof_id(tmp_id)

    # Test python int
    tmp_id = int(tof_mrpc_id_np[is_mrpc][0])
    tmp_endcap, tmp_module, tmp_strip, tmp_end = (
        det_id.tof_id_to_mrpc_endcap(tmp_id),
        det_id.tof_id_to_mrpc_module(tmp_id),
        det_id.tof_id_to_mrpc_strip(tmp_id),
        det_id.tof_id_to_end(tmp_id),
    )
    assert tmp_endcap == np_endcap[is_mrpc][0]
    assert tmp_module == np_module[is_mrpc][0]
    assert tmp_strip == np_strip[is_mrpc][0]
    assert tmp_end == np_end[is_mrpc][0]
    assert p3.get_tof_mrpc_id(tmp_endcap, tmp_module, tmp_strip, tmp_end) == tmp_id
    assert det_id.check_tof_id(tmp_id)


def test_muc_id():
    muc_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_mucDigiCol"
    ].array()["m_intId"]

    # Test awkward
    ak_part, ak_segment, ak_layer, ak_channel, ak_gap, ak_strip = (
        det_id.muc_id_to_part(muc_id_ak),
        det_id.muc_id_to_segment(muc_id_ak),
        det_id.muc_id_to_layer(muc_id_ak),
        det_id.muc_id_to_channel(muc_id_ak),
        det_id.muc_id_to_gap(muc_id_ak),
        det_id.muc_id_to_strip(muc_id_ak),
    )
    assert ak.all(p3.get_muc_id(ak_part, ak_segment, ak_layer, ak_channel) == muc_id_ak)
    assert ak.all(det_id.check_muc_id(muc_id_ak))
    assert ak.all(ak_layer == ak_gap)
    assert ak.all(ak_channel == ak_strip)

    # Test numpy
    muc_id_np = ak.flatten(muc_id_ak).to_numpy()
    np_part, np_segment, np_layer, np_channel, np_gap, np_strip = (
        det_id.muc_id_to_part(muc_id_np),
        det_id.muc_id_to_segment(muc_id_np),
        det_id.muc_id_to_layer(muc_id_np),
        det_id.muc_id_to_channel(muc_id_np),
        det_id.muc_id_to_gap(muc_id_np),
        det_id.muc_id_to_strip(muc_id_np),
    )
    assert np.all(p3.get_muc_id(np_part, np_segment, np_layer, np_channel) == muc_id_np)
    assert np.all(det_id.check_muc_id(muc_id_np))
    assert np.all(np_layer == np_gap)
    assert np.all(np_channel == np_strip)

    # Test uint32
    tmp_id = muc_id_np[0]
    tmp_part, tmp_segment, tmp_layer, tmp_channel, tmp_gap, tmp_strip = (
        det_id.muc_id_to_part(tmp_id),
        det_id.muc_id_to_segment(tmp_id),
        det_id.muc_id_to_layer(tmp_id),
        det_id.muc_id_to_channel(tmp_id),
        det_id.muc_id_to_gap(tmp_id),
        det_id.muc_id_to_strip(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_segment == np_segment[0]
    assert tmp_layer == np_layer[0]
    assert tmp_channel == np_channel[0]
    assert tmp_gap == np_gap[0]
    assert tmp_strip == np_strip[0]
    assert p3.get_muc_id(tmp_part, tmp_segment, tmp_layer, tmp_channel) == tmp_id
    assert det_id.check_muc_id(tmp_id)

    # Test python int
    tmp_id = int(muc_id_np[0])
    tmp_part, tmp_segment, tmp_layer, tmp_channel, tmp_gap, tmp_strip = (
        det_id.muc_id_to_part(tmp_id),
        det_id.muc_id_to_segment(tmp_id),
        det_id.muc_id_to_layer(tmp_id),
        det_id.muc_id_to_channel(tmp_id),
        det_id.muc_id_to_gap(tmp_id),
        det_id.muc_id_to_strip(tmp_id),
    )
    assert tmp_part == np_part[0]
    assert tmp_segment == np_segment[0]
    assert tmp_layer == np_layer[0]
    assert tmp_channel == np_channel[0]
    assert tmp_gap == np_gap[0]
    assert tmp_strip == np_strip[0]
    assert p3.get_muc_id(tmp_part, tmp_segment, tmp_layer, tmp_channel) == tmp_id
    assert det_id.check_muc_id(tmp_id)


def test_cgem_id():
    cgem_id_ak: ak.Array = p3.open(data_dir / "test_cgem.rtraw")[
        "Event/TDigiEvent/m_cgemDigiCol"
    ].array()["m_intId"]

    # Test awkward
    ak_layer, ak_sheet, ak_strip, ak_is_x_strip = (
        det_id.cgem_id_to_layer(cgem_id_ak),
        det_id.cgem_id_to_sheet(cgem_id_ak),
        det_id.cgem_id_to_strip(cgem_id_ak),
        det_id.cgem_id_to_is_x_strip(cgem_id_ak),
    )
    assert ak.all(p3.get_cgem_id(ak_layer, ak_sheet, ak_strip, ak_is_x_strip) == cgem_id_ak)
    assert ak.all(det_id.check_cgem_id(cgem_id_ak))

    # Test numpy
    cgem_id_np = ak.flatten(cgem_id_ak).to_numpy()
    np_layer, np_sheet, np_strip, np_is_x_strip = (
        det_id.cgem_id_to_layer(cgem_id_np),
        det_id.cgem_id_to_sheet(cgem_id_np),
        det_id.cgem_id_to_strip(cgem_id_np),
        det_id.cgem_id_to_is_x_strip(cgem_id_np),
    )
    assert np.all(p3.get_cgem_id(np_layer, np_sheet, np_strip, np_is_x_strip) == cgem_id_np)
    assert np.all(det_id.check_cgem_id(cgem_id_np))

    # Test uint32
    tmp_id = cgem_id_np[0]
    tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip = (
        det_id.cgem_id_to_layer(tmp_id),
        det_id.cgem_id_to_sheet(tmp_id),
        det_id.cgem_id_to_strip(tmp_id),
        det_id.cgem_id_to_is_x_strip(tmp_id),
    )
    assert tmp_layer == np_layer[0]
    assert tmp_sheet == np_sheet[0]
    assert tmp_strip == np_strip[0]
    assert tmp_is_x_strip == np_is_x_strip[0]
    assert p3.get_cgem_id(tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip) == tmp_id
    assert det_id.check_cgem_id(tmp_id)

    # Test python int
    tmp_id = int(cgem_id_np[0])
    tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip = (
        det_id.cgem_id_to_layer(tmp_id),
        det_id.cgem_id_to_sheet(tmp_id),
        det_id.cgem_id_to_strip(tmp_id),
        det_id.cgem_id_to_is_x_strip(tmp_id),
    )
    assert tmp_layer == np_layer[0]
    assert tmp_sheet == np_sheet[0]
    assert tmp_strip == np_strip[0]
    assert tmp_is_x_strip == np_is_x_strip[0]
    assert p3.get_cgem_id(tmp_layer, tmp_sheet, tmp_strip, tmp_is_x_strip) == tmp_id
    assert det_id.check_cgem_id(tmp_id)


def test_parse_mdc_id():
    mdc_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_mdcDigiCol"
    ].array()["m_intId"]

    wire_ak = det_id.mdc_id_to_wire(mdc_id_ak)
    layer_ak = det_id.mdc_id_to_layer(mdc_id_ak)
    is_stereo_ak = det_id.mdc_id_to_is_stereo(mdc_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_mdc_id(mdc_id_ak, flat=False, library="ak")
    assert ak_res1.fields == ["wire", "layer", "is_stereo"]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["wire"] == wire_ak)
    assert ak.all(ak_res1["layer"] == layer_ak)
    assert ak.all(ak_res1["is_stereo"] == is_stereo_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_mdc_id(mdc_id_ak, flat=True, library="ak")
    assert ak_res2.fields == ["wire", "layer", "is_stereo"]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["wire"] == ak.flatten(wire_ak))
    assert ak.all(ak_res2["layer"] == ak.flatten(layer_ak))
    assert ak.all(ak_res2["is_stereo"] == ak.flatten(is_stereo_ak))

    mdc_id_np = ak.flatten(mdc_id_ak).to_numpy()
    wire_np = ak.flatten(wire_ak).to_numpy()
    layer_np = ak.flatten(layer_ak).to_numpy()
    is_stereo_np = ak.flatten(is_stereo_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_mdc_id(mdc_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == ["wire", "layer", "is_stereo"]
    assert np.all(np_res1["wire"] == wire_np)
    assert np.all(np_res1["layer"] == layer_np)
    assert np.all(np_res1["is_stereo"] == is_stereo_np)

    # Test int, library='ak'
    mdc_id_int = int(mdc_id_np[0])
    int_res1 = p3.parse_mdc_id(mdc_id_int, flat=False, library="ak")
    assert int_res1.fields == ["wire", "layer", "is_stereo"]
    assert int_res1["wire"] == wire_np[0]
    assert int_res1["layer"] == layer_np[0]
    assert int_res1["is_stereo"] == is_stereo_np[0]

    # Test int, library='np'
    int_res2 = p3.parse_mdc_id(mdc_id_int, flat=False, library="np")
    assert list(int_res2.keys()) == ["wire", "layer", "is_stereo"]
    assert int_res2["wire"] == wire_np[0]
    assert int_res2["layer"] == layer_np[0]
    assert int_res2["is_stereo"] == is_stereo_np[0]


def test_parse_emc_id():
    emc_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_emcDigiCol"
    ].array()["m_intId"]

    module_ak = det_id.emc_id_to_module(emc_id_ak)
    theta_ak = det_id.emc_id_to_theta(emc_id_ak)
    phi_ak = det_id.emc_id_to_phi(emc_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_emc_id(emc_id_ak, flat=False, library="ak")
    assert ak_res1.fields == ["module", "theta", "phi"]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["module"] == module_ak)
    assert ak.all(ak_res1["theta"] == theta_ak)
    assert ak.all(ak_res1["phi"] == phi_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_emc_id(emc_id_ak, flat=True, library="ak")
    assert ak_res2.fields == ["module", "theta", "phi"]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["module"] == ak.flatten(module_ak))
    assert ak.all(ak_res2["theta"] == ak.flatten(theta_ak))
    assert ak.all(ak_res2["phi"] == ak.flatten(phi_ak))

    emc_id_np = ak.flatten(emc_id_ak).to_numpy()
    module_np = ak.flatten(module_ak).to_numpy()
    theta_np = ak.flatten(theta_ak).to_numpy()
    phi_np = ak.flatten(phi_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_emc_id(emc_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == ["module", "theta", "phi"]
    assert np.all(np_res1["module"] == module_np)
    assert np.all(np_res1["theta"] == theta_np)
    assert np.all(np_res1["phi"] == phi_np)

    # Test int, library='ak'
    emc_id_int = int(emc_id_np[0])
    int_res1 = p3.parse_emc_id(emc_id_int, flat=False, library="ak")
    assert int_res1.fields == ["module", "theta", "phi"]
    assert int_res1["module"] == module_np[0]
    assert int_res1["theta"] == theta_np[0]
    assert int_res1["phi"] == phi_np[0]

    # Test int, library='np'
    int_res2 = p3.parse_emc_id(emc_id_int, flat=False, library="np")
    assert list(int_res2.keys()) == ["module", "theta", "phi"]
    assert int_res2["module"] == module_np[0]
    assert int_res2["theta"] == theta_np[0]
    assert int_res2["phi"] == phi_np[0]


def test_parse_tof_id():
    tof_id_ak: ak.Array = p3.open(data_dir / "test_mrpc.rtraw")[
        "Event/TDigiEvent/m_tofDigiCol"
    ].array()["m_intId"]

    part_ak = det_id.tof_id_to_part(tof_id_ak)
    end_ak = det_id.tof_id_to_end(tof_id_ak)
    scint_layer_ak = det_id.tof_id_to_scint_layer(tof_id_ak)
    scint_phi_ak = det_id.tof_id_to_scint_phi(tof_id_ak)
    mrpc_endcap_ak = det_id.tof_id_to_mrpc_endcap(tof_id_ak)
    mrpc_module_ak = det_id.tof_id_to_mrpc_module(tof_id_ak)
    mrpc_strip_ak = det_id.tof_id_to_mrpc_strip(tof_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_tof_id(tof_id_ak, flat=False, library="ak")
    assert ak_res1.fields == [
        "part",
        "end",
        "scint_layer",
        "scint_phi",
        "mrpc_endcap",
        "mrpc_module",
        "mrpc_strip",
        "is_mrpc",
    ]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["part"] == part_ak)
    assert ak.all(ak_res1["end"] == end_ak)
    assert ak.all(ak_res1["scint_layer"] == scint_layer_ak)
    assert ak.all(ak_res1["scint_phi"] == scint_phi_ak)
    assert ak.all(ak_res1["mrpc_endcap"] == mrpc_endcap_ak)
    assert ak.all(ak_res1["mrpc_module"] == mrpc_module_ak)
    assert ak.all(ak_res1["mrpc_strip"] == mrpc_strip_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_tof_id(tof_id_ak, flat=True, library="ak")
    assert ak_res2.fields == [
        "part",
        "end",
        "scint_layer",
        "scint_phi",
        "mrpc_endcap",
        "mrpc_module",
        "mrpc_strip",
        "is_mrpc",
    ]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["part"] == ak.flatten(part_ak))
    assert ak.all(ak_res2["end"] == ak.flatten(end_ak))
    assert ak.all(ak_res2["scint_layer"] == ak.flatten(scint_layer_ak))
    assert ak.all(ak_res2["scint_phi"] == ak.flatten(scint_phi_ak))
    assert ak.all(ak_res2["mrpc_endcap"] == ak.flatten(mrpc_endcap_ak))
    assert ak.all(ak_res2["mrpc_module"] == ak.flatten(mrpc_module_ak))
    assert ak.all(ak_res2["mrpc_strip"] == ak.flatten(mrpc_strip_ak))

    tof_id_np = ak.flatten(tof_id_ak).to_numpy()
    part_np = ak.flatten(part_ak).to_numpy()
    end_np = ak.flatten(end_ak).to_numpy()
    scint_layer_np = ak.flatten(scint_layer_ak).to_numpy()
    scint_phi_np = ak.flatten(scint_phi_ak).to_numpy()
    mrpc_endcap_np = ak.flatten(mrpc_endcap_ak).to_numpy()
    mrpc_module_np = ak.flatten(mrpc_module_ak).to_numpy()
    mrpc_strip_np = ak.flatten(mrpc_strip_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_tof_id(tof_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == [
        "part",
        "end",
        "scint_layer",
        "scint_phi",
        "mrpc_endcap",
        "mrpc_module",
        "mrpc_strip",
        "is_mrpc",
    ]
    assert np.all(np_res1["part"] == part_np)
    assert np.all(np_res1["end"] == end_np)
    assert np.all(np_res1["scint_layer"] == scint_layer_np)
    assert np.all(np_res1["scint_phi"] == scint_phi_np)
    assert np.all(np_res1["mrpc_endcap"] == mrpc_endcap_np)
    assert np.all(np_res1["mrpc_module"] == mrpc_module_np)
    assert np.all(np_res1["mrpc_strip"] == mrpc_strip_np)

    # Test int, library='ak'
    tof_id_int = int(tof_id_np[0])
    int_res1 = p3.parse_tof_id(tof_id_int, flat=False, library="ak")
    assert int_res1.fields == [
        "part",
        "end",
        "scint_layer",
        "scint_phi",
        "mrpc_endcap",
        "mrpc_module",
        "mrpc_strip",
        "is_mrpc",
    ]
    assert int_res1["part"] == part_np[0]
    assert int_res1["end"] == end_np[0]
    assert int_res1["scint_layer"] == scint_layer_np[0]
    assert int_res1["scint_phi"] == scint_phi_np[0]
    assert int_res1["mrpc_endcap"] == mrpc_endcap_np[0]
    assert int_res1["mrpc_module"] == mrpc_module_np[0]
    assert int_res1["mrpc_strip"] == mrpc_strip_np[0]

    # Test int, library='np'
    int_res2 = p3.parse_tof_id(tof_id_int, flat=False, library="np")
    assert list(int_res2.keys()) == [
        "part",
        "end",
        "scint_layer",
        "scint_phi",
        "mrpc_endcap",
        "mrpc_module",
        "mrpc_strip",
        "is_mrpc",
    ]
    assert int_res2["part"] == part_np[0]
    assert int_res2["end"] == end_np[0]
    assert int_res2["scint_layer"] == scint_layer_np[0]
    assert int_res2["scint_phi"] == scint_phi_np[0]
    assert int_res2["mrpc_endcap"] == mrpc_endcap_np[0]
    assert int_res2["mrpc_module"] == mrpc_module_np[0]
    assert int_res2["mrpc_strip"] == mrpc_strip_np[0]


def test_parse_muc_id():
    muc_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_mucDigiCol"
    ].array()["m_intId"]

    part_ak = det_id.muc_id_to_part(muc_id_ak)
    segment_ak = det_id.muc_id_to_segment(muc_id_ak)
    layer_ak = det_id.muc_id_to_layer(muc_id_ak)
    channel_ak = det_id.muc_id_to_channel(muc_id_ak)
    gap_ak = det_id.muc_id_to_gap(muc_id_ak)
    strip_ak = det_id.muc_id_to_strip(muc_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_muc_id(muc_id_ak, flat=False, library="ak")
    assert ak_res1.fields == [
        "part",
        "segment",
        "layer",
        "channel",
        "gap",
        "strip",
    ]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["part"] == part_ak)
    assert ak.all(ak_res1["segment"] == segment_ak)
    assert ak.all(ak_res1["layer"] == layer_ak)
    assert ak.all(ak_res1["channel"] == channel_ak)
    assert ak.all(ak_res1["gap"] == gap_ak)
    assert ak.all(ak_res1["strip"] == strip_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_muc_id(muc_id_ak, flat=True, library="ak")
    assert ak_res2.fields == [
        "part",
        "segment",
        "layer",
        "channel",
        "gap",
        "strip",
    ]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["part"] == ak.flatten(part_ak))
    assert ak.all(ak_res2["segment"] == ak.flatten(segment_ak))
    assert ak.all(ak_res2["layer"] == ak.flatten(layer_ak))
    assert ak.all(ak_res2["channel"] == ak.flatten(channel_ak))
    assert ak.all(ak_res2["gap"] == ak.flatten(gap_ak))
    assert ak.all(ak_res2["strip"] == ak.flatten(strip_ak))

    muc_id_np = ak.flatten(muc_id_ak).to_numpy()
    part_np = ak.flatten(part_ak).to_numpy()
    segment_np = ak.flatten(segment_ak).to_numpy()
    layer_np = ak.flatten(layer_ak).to_numpy()
    channel_np = ak.flatten(channel_ak).to_numpy()
    gap_np = ak.flatten(gap_ak).to_numpy()
    strip_np = ak.flatten(strip_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_muc_id(muc_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == [
        "part",
        "segment",
        "layer",
        "channel",
        "gap",
        "strip",
    ]
    assert np.all(np_res1["part"] == part_np)
    assert np.all(np_res1["segment"] == segment_np)
    assert np.all(np_res1["layer"] == layer_np)
    assert np.all(np_res1["channel"] == channel_np)
    assert np.all(np_res1["gap"] == gap_np)
    assert np.all(np_res1["strip"] == strip_np)


def test_parse_cgem_id():
    cgem_id_ak: ak.Array = p3.open(data_dir / "test_cgem.rtraw")[
        "Event/TDigiEvent/m_cgemDigiCol"
    ].array()["m_intId"]

    layer_ak = det_id.cgem_id_to_layer(cgem_id_ak)
    sheet_ak = det_id.cgem_id_to_sheet(cgem_id_ak)
    strip_ak = det_id.cgem_id_to_strip(cgem_id_ak)
    is_x_strip_ak = det_id.cgem_id_to_is_x_strip(cgem_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_cgem_id(cgem_id_ak, flat=False, library="ak")
    assert ak_res1.fields == ["layer", "sheet", "strip", "is_x_strip"]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["layer"] == layer_ak)
    assert ak.all(ak_res1["sheet"] == sheet_ak)
    assert ak.all(ak_res1["strip"] == strip_ak)
    assert ak.all(ak_res1["is_x_strip"] == is_x_strip_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_cgem_id(cgem_id_ak, flat=True, library="ak")
    assert ak_res2.fields == ["layer", "sheet", "strip", "is_x_strip"]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["layer"] == ak.flatten(layer_ak))
    assert ak.all(ak_res2["sheet"] == ak.flatten(sheet_ak))
    assert ak.all(ak_res2["strip"] == ak.flatten(strip_ak))
    assert ak.all(ak_res2["is_x_strip"] == ak.flatten(is_x_strip_ak))

    cgem_id_np = ak.flatten(cgem_id_ak).to_numpy()
    layer_np = ak.flatten(layer_ak).to_numpy()
    strip_np = ak.flatten(strip_ak).to_numpy()
    sheet_np = ak.flatten(sheet_ak).to_numpy()
    is_x_strip_np = ak.flatten(is_x_strip_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_cgem_id(cgem_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == ["layer", "sheet", "strip", "is_x_strip"]
    assert np.all(np_res1["layer"] == layer_np)
    assert np.all(np_res1["sheet"] == sheet_np)
    assert np.all(np_res1["strip"] == strip_np)
    assert np.all(np_res1["is_x_strip"] == is_x_strip_np)

    # Test int, library='ak'
    cgem_id_int = int(cgem_id_np[0])
    int_res1 = p3.parse_cgem_id(cgem_id_int, flat=False, library="ak")
    assert int_res1.fields == ["layer", "sheet", "strip", "is_x_strip"]
    assert int_res1["layer"] == layer_np[0]
    assert int_res1["sheet"] == sheet_np[0]
    assert int_res1["strip"] == strip_np[0]
    assert int_res1["is_x_strip"] == is_x_strip_np[0]

    # Test int, library='np'
    int_res2 = p3.parse_cgem_id(cgem_id_int, flat=False, library="np")
    assert list(int_res2.keys()) == ["layer", "sheet", "strip", "is_x_strip"]
    assert int_res2["layer"] == layer_np[0]
    assert int_res2["sheet"] == sheet_np[0]
    assert int_res2["strip"] == strip_np[0]
    assert int_res2["is_x_strip"] == is_x_strip_np[0]



from . import identify
from .identify import (
    get_cgem_id,
    get_emc_id,
    get_mdc_id,
    get_muc_id,
    get_tof_mrpc_id,
    get_tof_scint_id,
    parse_cgem_id,
    parse_emc_id,
    parse_mdc_id,
    parse_muc_id,
    parse_tof_id,
)
